from workspaces.capability.enums import (
    CapabilityEventType,
    CapabilityStepType,
    ExecutionPriority,
    ExecutionState,
)
from workspaces.capability.helpers import CapabilitySequence
from workspaces.capability.schema import CapabilityEvent, CapabilityRequest
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.capability_queue import CapabilityQueue
from workspaces.capability.services.interfaces import CapabilityInfoIF
from workspaces.workflow.services.interfaces import WorkflowServiceIF


class ExecutionManager:
    """
    Class that manages the capability execution system. It's duties include:
    - Creating new executions
    - Managing existing execution state
    - Handling capability events and updating the state of the system accordingly
    """

    def __init__(self, capability_info: CapabilityInfoIF, workflow_service: WorkflowServiceIF):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = capability_info
        self.workflow_service = workflow_service

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings
        TODO: Executions should be made from capability versions, not requests

        :param request: Capability request
        :return: Execution of the request
        """
        execution = self.capability_info.create_execution(request)
        steps = CapabilitySequence.from_str(execution.steps)

        if steps[0].step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # First step is to run workflow, move to capability queue
            # FIXME: Priority needs to be dynamic
            self.enqueue_execution(execution, ExecutionPriority.Default.value)
        else:
            self.execution_pool.append(execution)
        return execution

    def update_execution(self, capability_event: CapabilityEvent):
        """
        Update capability execution given a received capability event

        :param capability_event: Received capability event
        """
        execution = self.capability_info.lookup_execution(capability_event.execution_id)
        step_sequence = CapabilitySequence.from_str(execution.steps)
        current_step = step_sequence[execution.current_step]
        event_type = CapabilityEventType[capability_event.event_type.name]

        # Check if event corresponds to current capability step
        if current_step.step_type.value == event_type.value:
            # Check if previous step (that just completed) is run workflow step
            if current_step.step_type == CapabilityStepType.PrepareAndRunWorkflow:
                if self.queues:
                    # Return capability engine to available state
                    self.queues.get(execution.capability.name).complete_execution(
                        capability_event.execution_id
                    )

            if not execution.on_last_step():
                execution.current_step += 1
            else:
                # Execution on its last step and received the proper event
                execution.state = ExecutionState.Complete.name
                self.capability_info.save_execution(execution)
                return

            # Check if upcoming step is run workflow step
            if current_step.step_type == CapabilityStepType.PrepareAndRunWorkflow:
                # Enqueue execution that is on a run workflow step
                # FIXME: Priority needs to be dynamic; perhaps a priority column in execution schema?
                self.enqueue_execution(execution, ExecutionPriority.Default.value)

            # Update execution record's state
            # execution.state = ExecutionState.Ready.name
            self.capability_info.save_execution(execution)
        else:
            # Event type does not correspond to current step
            # TODO: Logging
            print(
                f"Mismatched event type {event_type} for execution {capability_event.execution_id}"
            )

    def enqueue_execution(
        self,
        execution: CapabilityExecutionIF,
        priority: int = ExecutionPriority.Default.value,
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue

        :param execution: Capability execution record
        :param priority: Priority of the requested execution
        """
        # Get correct queue or initialize one
        queue = self.queues.get(
            execution.capability.name,
            CapabilityQueue(
                self.capability_info,
                self.workflow_service,
                execution.capability.max_jobs,
            ),
        )
        queue.enqueue(execution, priority)
        self.queues[execution.capability.name] = queue

        # Remove execution record from pool
        if execution in self.execution_pool:
            self.execution_pool.remove(execution)
